#include <PalmOS.h>



struct ProtectedDb {
	UInt16 lockCt;
	UInt16 cardNo;
	LocalID LID;
};

Err pDmDeleteDatabase(MemHandle protectedListH, Err (*oldTrap)(UInt16 cardNo, LocalID LID), void *retAddr, UInt16 cardNo, LocalID LID)
{
	struct ProtectedDb *dbL;
	Err e = errNone;
	UInt32 i, num;

	
	dbL = MemHandleLock(protectedListH);
	num = MemHandleSize(protectedListH) / sizeof(struct ProtectedDb);
	for (i = 0; i < num; i++) {
		if (dbL[i].cardNo == cardNo && dbL[i].LID == LID) {
			e = dmErrDatabaseProtected;
			break;
		}
	}
	MemHandleUnlock(protectedListH);
	
	if (e == errNone)
		e = oldTrap(cardNo, LID);
	
	return e;
}

Err pDmDatabaseProtect(MemHandle protectedListH, void* retAddr, UInt16 cardNo, LocalID LID, Boolean protect)
{
	struct ProtectedDb *dbL;
	UInt32 i, sz, num;
	Err e = errNone;
	
	
	if (MemLocalIDKind(LID) != memIDHandle)
		return dmErrROMBased;
	
	dbL = MemHandleLock(protectedListH);
	num = MemHandleSize(protectedListH) / sizeof(struct ProtectedDb);
	for (i = 0; i < num; i++) {
		if (dbL[i].cardNo == cardNo && dbL[i].LID == LID) {
			if (protect) {
				dbL[i].lockCt++;
				if (!dbL[i].lockCt)
					SysFatalAlert("DB overprotected");
			}
			else if (!--dbL[i].lockCt) {
				dbL[i] = dbL[--num];
				(void)MemHandleResize(protectedListH, num ? num * sizeof(struct ProtectedDb) : 1);	//cannot fail on shrinking
			}
			goto out;
		}
	}
	
	
	if (!protect)
		e = dmErrDatabaseNotProtected;
	else {
	
		//we'll need to allocate a new slot
		MemHandleUnlock(protectedListH);
	
		e = MemHandleResize(protectedListH, (num + 1) * sizeof(struct ProtectedDb));
		if (e != errNone)
			goto out_no_unlock;
		
		dbL = MemHandleLock(protectedListH);
		dbL[num].lockCt = 1;
		dbL[num].cardNo = cardNo;
		dbL[num].LID = LID;
	}
	
out:
	MemHandleUnlock(protectedListH);

out_no_unlock:
	return e;
}

#ifdef __GNUC__
	
	static void pDmDatabaseProtectWrapper(void)
	{
		//with no "naked" attribute available, this is a bit fragile, but as long as you do not try to compile with -O0, it will be ok
		
		asm volatile(
			"	move.l #0xaaaaaaaa, -(%sp)			\n\t"		//protected list handle
			"	bsr    pDmDatabaseProtect			\n\t"
			"	lea    4(%sp), %sp					\n\t"
		);
	}
	
	static void pDmDeleteDatabaseWrapper(void)
	{
		//with no "naked" attribute available, this is a bit fragile, but as long as you do not try to compile with -O0, it will be ok
		//also gcc gets too clever with tailcalls, so to avoid them, we provide a proper stack
		//sadly this means we need to create pointless stack frames
		//c'est la vie
		
		asm volatile(
			"	move.l  #0xaaaaaaaa, %d0			\n\t"		//old trap
			"	movea.l  #0xaaaaaaaa, %a0			\n\t"		//protected list handle
			"	move.l  6(%sp), -(%sp)				\n\t"
			"	move.w  8(%sp), -(%sp)				\n\t"
			"	clr.l   -(%sp)						\n\t"
			"	move.l  %d0, -(%sp)					\n\t"
			"	move.l  %a0, -(%sp)					\n\t"
			"	bsr    pDmDeleteDatabase			\n\t"
			"	lea    18(%sp), %sp					\n\t"
		);
	}
	
#else

	asm void pDmDatabaseProtectWrapper(void)
	{
		move.l #0xaaaaaaaa, -(sp)			//protected list handle
		jsr    pDmDatabaseProtect
		lea    4(sp), sp
		rts
	}
	
	asm void pDmDeleteDatabaseWrapper(void)
	{
		move.l #0xaaaaaaaa, -(sp)			//old trap
		move.l #0xaaaaaaaa, -(sp)			//protected list handle
		jsr    pDmDeleteDatabase
		lea    8(sp), sp
		rts
	}
#endif

void __Startup__(void)
{
	UInt32 ver;
		
	if (errNone == FtrGet(sysFtrCreator, sysFtrNumROMVersion, &ver) && ver < sysMakeROMVersion(2, 0, 0, 0, 0)) {
		
		MemHandle listH = MemChunkNew(0, 1, 0);
		
		MemSemaphoreReserve(true);
		*(UInt32*)(((UInt16*)pDmDatabaseProtectWrapper) + 1) = (UInt32)listH;
		*(UInt32*)(((UInt16*)pDmDeleteDatabaseWrapper) + 1) = (UInt32)SysGetTrapAddress(sysTrapDmDeleteDatabase);
		*(UInt32*)(((UInt16*)pDmDeleteDatabaseWrapper) + 4) = (UInt32)listH;
		MemSemaphoreRelease(true);
		
		SysSetTrapAddress(sysTrapDmDeleteDatabase, pDmDeleteDatabaseWrapper);
		SysSetTrapAddress(sysTrapDmDatabaseProtect, pDmDatabaseProtectWrapper);
		
		MemHandleLock(DmGet1Resource('extn', 2));
	}
	
	//do this here in case we DID just create the trap
	(void)DmDatabaseProtect(0, DmFindDatabase(0, "ComboPatch"), true);
}
