#include <PalmOS.h>
#include <stdarg.h>


static Int16 PrvStrCompare(const Char* s1, const Char* s2, Int32 n, Boolean caseless)
{
	UInt32 len1 = StrLen(s1), len2 = StrLen(s2);
	UInt32 dummy;
	
	//verify we're not ending mid-char
	if (n < (Int32)len1)
		TxtCharBounds(s1, n, &len1, &dummy);
	
	if (n < (Int32)len2)
		TxtCharBounds(s2, n, &len2, &dummy);
	
	if (caseless)
		return TxtCaselessCompare(s1, len1, NULL, s2, len2, NULL);
	else
		return TxtCompare(s1, len1, NULL, s2, len2, NULL);
}

Int16 pStrNCompare(const Char* s1, const Char* s2, Int32 n)
{
	return PrvStrCompare(s1, s2, n, false);
}

Int16 pStrNCaselessCompare(const Char* s1, const Char* s2, Int32 n)
{
	return PrvStrCompare(s1, s2, n, true);
}

static char* pStrNCopy(char* dstP, const char* srcP, Int16 n)
{
	const char *src = srcP;
	char *dst = dstP;
	char ch;
	
	if (!src || !dst || n <= 0)
		return dstP;
	
	//copy up to N chars, while also respecting string end
	while (n && (ch = *src)) {
		
		n--;
		*dst++ = ch;
		src++;
	}
	
	if (n)
		*dst = 0;
	else {			//verify we didnt copy half a character, if we did, zero-term before going that far
		
		UInt32 now = src - srcP, chrBegin, chrEnd;
		
		TxtCharBounds(srcP, now, &chrBegin, &chrEnd);
		if (chrBegin < now)
			dstP[chrBegin] = 0;
	}
	return dstP;
}

static Int16 pStrPrintF(char *dst, const char* fmt, ...)
{
	va_list vl;
	Int16 ret;
	

	va_start(vl, fmt);
	ret = StrVPrintF(dst, fmt, vl);
	va_end(vl);
	
	return ret;
}


#define FLAGS_LEFT_JUSTIFY			0x01
#define FLAGS_ZERO_PAD				0x02
#define FLAGS_IS_LONG				0x04

static void printfPrvPrintNumber(char *dstP, UInt32 val)
{
	char *dst = dstP;
	
	//make it in reverse
	do {
		*dst++ = '0' + val % 10;
	} while (val /= 10);
	
	//terminate and repoint to last printed char
	*dst-- = 0;
	
	//now reverse it
	while (dst > dstP) {
		
		char ch = *dst;
		*dst = *dstP;
		*dstP = ch;
		
		dst--;
		dstP++;
	}
}

static void printfPrvPrintHexNum(char *dstP, UInt32 val, UInt8 flags)
{
	char *dst = dstP + ((flags & FLAGS_IS_LONG) ? 8 : 4);
	
	//terminate
	*dst = 0;
	
	do {
		
		char ch = val & 0x0f;
		val >>= 4;
		
		*--dst = ch + ((ch >= 10) ? 'A' - 10 : '0');
		
	} while (dst != dstP);
}

static char* printfPrvAppendWithPadding(char *dst, char *src, Int16 minLen, char sign, UInt8 flags)		//no need to zero-terminate
{
	char ch, pad = (flags & FLAGS_ZERO_PAD) ? '0' : ' ';
	Int16 strLen = StrLen(src) + (sign ? 1 : 0);
	
	//maybe we do not need padding at all? or if we do, it is on the right of the string?
	if (strLen >= minLen || (flags & FLAGS_LEFT_JUSTIFY)) {
		
		//sign
		if (sign)
			*dst++ = sign;
		
		//string
		while ((ch = *src++) != 0)
			*dst++ = ch;
		
		//pad
		while (strLen++ < minLen)
			*dst++ = pad;
	}
	else if (flags & FLAGS_ZERO_PAD) {		//zero padding, right aligned
		
		//sign
		if (sign)
			*dst++ = sign;
		
		//pad
		while (strLen++ < minLen)
			*dst++ = pad;
		
		//string
		while ((ch = *src++) != 0)
			*dst++ = ch;
	}
	else {									//space-padding, right aligned
		
		//pad
		while (strLen++ < minLen)
			*dst++ = pad;
		//sign
		if (sign)
			*dst++ = sign;
		
		//string
		while ((ch = *src++) != 0)
			*dst++ = ch;
		
	}
	
	return dst;
}

static Int16 pStrVPrintF(char *dstP, const char* fmt, va_list vl)	//only implements what PalmOS does (not quite POSIX compliant, hehe)
{
	char ch, *dst = dstP;
	
	while ((ch = *fmt++) != 0) {
		
		if (ch != '%')
			*dst++ = ch;
		else {
			
			char tmpBuf[12], *strP = tmpBuf, sign = 0;
			Int16 minLen = 0;
			UInt8 flags = 0;
			UInt32 val;
			Int32 sval;
			
			//padding
			while (1)  switch (ch = *fmt) {
					
				case '+':
					sign = '+';
					fmt++;
					break;
				
				case ' ':
					if (!sign)
						sign = ' ';
					fmt++;
					break;
				
				case '-':
					flags |= FLAGS_LEFT_JUSTIFY;
					fmt++;
					break;
				
				case '0':
					flags |= FLAGS_ZERO_PAD;
					fmt++;
					break;
				
				default:
					goto padding_understood;
			}
padding_understood:
			
			//min width
			if (*fmt == '*') {
				//provided by Int16 in args
				minLen = va_arg(vl, int);
				fmt++;
			}
			else while ((ch = *fmt) >= '0' && ch <= '9') {
				
				minLen = minLen * 10 + ch - '0';
				fmt++;
			}

			//int size 
			switch (*fmt) {
				case 'h':	//short, which for us means nothing due to Palm's type size choice
					fmt++;
					break;
				
				case 'l':
					flags |= FLAGS_IS_LONG;
					fmt++;
					break;
			}
			
			//what is it?
			switch (ch = *fmt++) {
				case '%':
				default:
					*dst++ = ch;
					fmt++;
					break;
				
				case 's':	//string
					strP = va_arg(vl, char*);
					goto print_with_padding;
				
				case 'c':	//char
					tmpBuf[0] = va_arg(vl, int);
					tmpBuf[1] = 0;
					goto print_with_padding;
				
				case 'C':	//wide char
					tmpBuf[TxtSetNextChar(tmpBuf, 0, (WChar)va_arg(vl, unsigned /* promotion */))] = 0;
					goto print_with_padding;
				
				case 'd':
				case 'i':
					if (flags & FLAGS_IS_LONG)
						sval = va_arg(vl, Int32);
					else
						sval = (Int16)va_arg(vl, int /* promotion */);
					
					if (sval < 0) {
						sval = -sval;
						sign = '-';
					}
					printfPrvPrintNumber(tmpBuf, sval);
					goto print_with_padding;
					
				case 'u':
					if (flags & FLAGS_IS_LONG)
						val = va_arg(vl, UInt32);
					else
						val = (UInt16)va_arg(vl, unsigned /* promotion */);
					
					printfPrvPrintNumber(tmpBuf, val);
					goto print_with_padding;
					
				case 'x':
				case 'X':
					if (flags & FLAGS_IS_LONG)
						val = va_arg(vl, UInt32);
					else
						val = (UInt16)va_arg(vl, unsigned /* promotion */);
					
					printfPrvPrintHexNum(tmpBuf, val, flags);
					
			print_with_padding:
					dst = printfPrvAppendWithPadding(dst, strP, minLen, sign, flags);
					break;
			}
		}
	}
	*dst++ = 0;	//zero-terminate
		
	return dst - dstP;
}

static Int16 pMemCmp(const void *s1, const void *s2, Int32 numBytes)
{
	const UInt8 *s1p = s1, *s2p = s2;
	Int16 result;
	
	while (numBytes-- > 0 && !(result = *s1p++ - *s2p++));
	
	return result;
}

void __Startup__(void)
{
	UInt32 ver;
	
	if (errNone != FtrGet(sysFtrCreator, sysFtrNumROMVersion, &ver))
		return;
	
	if (ver < sysMakeROMVersion(2, 0, 0, 0, 0)) {
		
		SysSetTrapAddress(sysTrapStrNCompare, pStrNCompare);
		SysSetTrapAddress(sysTrapStrNCaselessCompare, pStrNCaselessCompare);
		SysSetTrapAddress(sysTrapMemCmp, pMemCmp);
	}
	
	if (ver < sysMakeROMVersion(3, 0, 0, 0, 0)) {
		
		SysSetTrapAddress(sysTrapStrNCopy, pStrNCopy);
		SysSetTrapAddress(sysTrapStrPrintF, pStrPrintF);
		SysSetTrapAddress(sysTrapStrVPrintF, pStrVPrintF);
		
		MemHandleLock(DmGet1Resource('extn', 8));
	}
}
