#include <PalmOS.h>
#include <ExpansionMgr.h>
#include <SlotDrvrLib.h>
#include <VFSMgr.h>
#include "ff.h"
#include "diskio.h"

#define MY_CRID							'fatf'


#define	BUILDING_APP_OR_LIB		1
#include <FSLib.h>
#undef BUILDING_APP_OR_LIB



#define CACHE_NUM_SECTORS		(32)		//only used at create time. at runtime actual number is used (if less than this)
#define OP_AT_ONCE				(16384)	//must fit in "int"





//CW SDK lacks this, so define it just in case
#ifndef expCapabilityHidden
	#define expCapabilityHidden         0x00000010  // card is hidden
#endif

struct VolumeState;
struct Globals;
static void cachePrvInit(struct Globals *g);
static void cachePrvClean(struct Globals *g, Boolean andInvalidate);


//Oppan Sony Style ...
	struct CardPartitionLayoutV1 {			//allocated by us with owner of zero. freed by caller
		UInt32 version;						// 0x00 - the value 1 is is stored here
		UInt32 partStartMinus1;				// 0x04 - if part is to start at sector 0x19, then 0x18 will be stored here
	};

	struct CardMetricsTypeV1 {				//used by expansion manager v1 (sony 3.5)
		UInt32 version;						// 0x00 - the value 1 is is stored here
		UInt32 totalSectors;				// 0x04
		UInt16 bytesPerSector;				// 0x08
		UInt8 sectorsPerBlock;				// 0x0a
		UInt8 rfu;							// 0x0b - the value 0 is written here
		UInt16 sectorsPerHead;				// 0x0c
		UInt16 headsPerCylinder;			// 0x0e
		UInt32 unknown_1;					// 0x10 - the value 1 is is stored here
		struct CardPartitionLayoutV1 *pl;	// 0x14 - allocated by the driver and freed by called. 8 bytes
	};


union FatTime {					//assumes BE system, allocating top to bottom bits
	UInt32 val;
	struct {
		UInt32 year		: 7;	//1980 - based
		UInt32 month	: 4;	//1 - based
		UInt32 day		: 5;	//1-based
		UInt32 hour		: 5;	//0-based
		UInt32 minute	: 6;	//0-based
		UInt32 secDiv2	: 5;	//0-based
	} part;
};


struct FileState {
	
	struct FileState *next, *prev;
	
	union {
		FileRef fileRef;
		struct {
			UInt32 isDir	: 1;
			UInt32 one		: 1;
			UInt32 seq		: 30;
		};
	};
	
	union {
		FIL f;
		DIR d;
	};
};

struct Globals;

struct CacheEntry {
	struct CacheEntry *next;
	struct CacheEntry *prev;
	UInt16 slotLibRef, slotRef;
	LBA_t sector;
	UInt8 dirty		: 1;
	UInt8 valid		: 1;
	MemHandle dataH;
};

struct VolumeState {
	struct VolumeState *next, *prev;
	struct Globals *g;
	UInt16 slotLibRef, slotRef, volRefNum;
	FATFS vol;
};

struct Globals {

	UInt16 refCnt;
	UInt32 nextFileRef;
	
	struct FileState *files;
	struct VolumeState *volumes;

#if CACHE_NUM_SECTORS
	struct CacheEntry *head, *tail;
	struct CacheEntry cacheEntries[CACHE_NUM_SECTORS];
	UInt16 numCacheEntries;
#endif

	//for temp use
	ExpCardInfoType ci;
	union {
		FIL tmpF;
		FILINFO tmpFI;
		UInt8 dataWriteBuf[slotSectorSize];
		TCHAR lblBuf[256];
	};
};

static const Err mErrMap[] = {
	[FR_OK] = errNone,
	[FR_DISK_ERR] = expErrCardBadSector,
	[FR_INT_ERR] = vfsErrFileGeneric,
	[FR_NOT_READY] = expErrCardNotPresent,
	[FR_NO_FILE] = vfsErrFileNotFound,
	[FR_NO_PATH] = vfsErrFileNotFound,
	[FR_INVALID_NAME] = vfsErrBadName,
	[FR_DENIED] = expErrCardReadOnly,
	[FR_EXIST] = vfsErrFileAlreadyExists,
	[FR_INVALID_OBJECT] = vfsErrFileGeneric,
	[FR_WRITE_PROTECTED] = expErrCardReadOnly,
	[FR_INVALID_DRIVE] = vfsErrVolumeBadRef,
	[FR_NOT_ENABLED] = vfsErrFileGeneric,
	[FR_NO_FILESYSTEM] = vfsErrBadData,		//i kid you not
	[FR_MKFS_ABORTED] = vfsErrNoFileSystem,
	[FR_TIMEOUT] = vfsErrFileGeneric,
	[FR_LOCKED] = vfsErrFilePermissionDenied,
	[FR_NOT_ENOUGH_CORE] = memErrNotEnoughSpace, 
	[FR_TOO_MANY_OPEN_FILES] = memErrNotEnoughSpace,
	[FR_INVALID_PARAMETER] = sysErrParamErr,
};






static Boolean isSonyEarlyVFS(void)	//PalmOS 3.5 Sony devices have an early implementation of ExpMgr/VFSMgr that differs in a few ways that we care about. This funcs tells us if this is such a device
{
	UInt32 expnVer;
	
	return errNone == FtrGet(sysFileCExpansionMgr, expFtrIDVersion, &expnVer) && expnVer == 1;
}

static struct Globals* fsPrvGetGlobals(UInt16 fsLibRefNum)
{
	SysLibTblEntryPtr entry = SysLibTblEntry(fsLibRefNum);
	struct Globals *g = (struct Globals*)entry->globalsP;
	
	if (!g || !g->refCnt)
		SysFatalAlert("NOT OPEN");
	
	return g;
}


static UInt32 fsPrvPalmTimeToFatTime(UInt32 palmTime)
{
	DateTimeType dtt;
	union FatTime time;
	
	TimSecondsToDateTime(palmTime, &dtt);

	if (dtt.year < 1980)
		dtt.year = 1980;
	
	time.part.year = dtt.year - 1980;
	time.part.month = dtt.month;
	time.part.day = dtt.day;
	time.part.hour = dtt.hour;
	time.part.minute = dtt.minute;
	time.part.secDiv2 = dtt.second / 2;
	
	return time.val;
}

static UInt32 fsPrvFatTimeToPalmTime(UInt32 fatTime)
{
	DateTimeType dtt;
	union FatTime time;
	
	time.val = fatTime;
	
	dtt.year = time.part.year + 1980;
	dtt.month = time.part.month;
	dtt.day = time.part.day;
	dtt.hour = time.part.hour;
	dtt.minute = time.part.minute;
	dtt.second  = time.part.secDiv2 * 2;
	
	return TimDateTimeToSeconds(&dtt);
}

static struct VolumeState* fsPrvFindVolume(struct Globals *g, UInt16 volRefNum)
{
	struct VolumeState *v;
	
	for (v = g->volumes; v; v = v->next) {
	
		if (v->volRefNum == volRefNum)
			return v;
	}
	
	return NULL;
}

static struct FileState* fsPrvFindFile(struct Globals *g, FileRef fileRef)
{
	struct FileState *f;
	
	for (f = g->files; f; f = f->next) {
	
		if (f->fileRef == fileRef)
			return f;
	}
	
	return NULL;
}


Err FSLibOpen(UInt16 fsLibRefNum)
{
	SysLibTblEntryPtr entry = SysLibTblEntry(fsLibRefNum);
	struct Globals *g = (struct Globals*)entry->globalsP;
	
	if (!g) {
		
		entry->globalsP = g = (struct Globals*)MemChunkNew(0, sizeof(struct Globals), memNewChunkFlagNonMovable);
		MemSet(g, sizeof(struct Globals), 0);
		
		cachePrvInit(g);
	}
	
	g->refCnt++;
		
	return errNone;
}

Err FSLibClose(UInt16 fsLibRefNum)
{
	SysLibTblEntryPtr entry = SysLibTblEntry(fsLibRefNum);
	struct Globals *g = (struct Globals*)entry->globalsP;
		
	if (--g->refCnt)
		return expErrStillOpen;
	
	entry->globalsP = NULL;
	
	return errNone;
}

Err FSLibSleep(UInt16 fsLibRefNum)
{
	return errNone;
}
	
Err FSLibWake(UInt16 fsLibRefNum)
{
	return errNone;
}

UInt32 FSLibAPIVersion(UInt16 fsLibRefNum)
{
	return isSonyEarlyVFS() ? 1 : fsLibAPIVersion;
}

Err FSCustomControl(UInt16 fsLibRefNum, UInt32 apiCreator, UInt16 apiSelector, void *valueP, UInt16 *valueLenP)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);

	return vfsErrUnimplemented;
}

Err FSFilesystemType(UInt16 fsLibRefNum, UInt32 *filesystemTypeP)
{
	if (filesystemTypeP)
		*filesystemTypeP = vfsFilesystemType_VFAT;
	
	return errNone;
}

Err FSFileCreate(UInt16 fsLibRefNum, UInt16 volRefNum, const Char *pathNameP)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct VolumeState *v = fsPrvFindVolume(g, volRefNum);
	FRESULT res;
	
	if (!v)
		return vfsErrVolumeBadRef;
	
	res = f_open(&v->vol, &g->tmpF, pathNameP, FA_WRITE | FA_CREATE_NEW);	/* Open or create a file */
	if (res == FR_OK)
		(void)f_close(&g->tmpF);
	
	cachePrvClean(g, false);
	
	return mErrMap[res];
}

Err FSFileOpen(UInt16 fsLibRefNum, UInt16 volRefNum, const Char *pathNameP, UInt16 openMode, FileRef *fileRefP)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct VolumeState *v = fsPrvFindVolume(g, volRefNum);
	struct FileState *f;
	FRESULT res;
	
	if (!v)
		return vfsErrVolumeBadRef;
	
	res = f_stat(&v->vol, pathNameP, &g->tmpFI);
	if (res != FR_OK)
		return mErrMap[res];
	
	f = MemChunkNew(0, sizeof(struct FileState), memNewChunkFlagNonMovable);
	if (!f)
		return memErrNotEnoughSpace;
	MemSet(f, sizeof(struct FileState), 0);

	if (g->tmpFI.fattrib & AM_DIR) {
				
		res = f_opendir(&v->vol, &f->d, pathNameP);
		f->isDir = true;
	}
	else {
		
		BYTE mode = 0;

		if (openMode & vfsModeRead)
			mode |= FA_READ;
		if (openMode & vfsModeWrite)
			mode |= FA_WRITE;
		
		res = f_open(&v->vol, &f->f, pathNameP, mode);
		f->isDir = false;
	}
		
	if (res != FR_OK)
		MemChunkFree(f);
	else {
		
		f->one = 1;
		f->seq = g->nextFileRef;
		
		while (fsPrvFindFile(g, f->fileRef)) 
			f->seq++;
		
		g->nextFileRef = f->seq + 1;
		
		if (g->files)
			g->files->prev = f;
		f->next = g->files;
		f->prev = NULL;
		g->files = f;
		
		*fileRefP = f->fileRef;
	}
	
	cachePrvClean(g, false);
	
	return mErrMap[res];
}

static FRESULT fsPrvFileClose(struct Globals *g, struct FileState *f)
{
	FRESULT res;
	
	if (f->isDir)
		res = f_closedir(&f->d);
	else
		res = f_close(&f->f);

	if (f->prev)
		f->prev->next = f->next;
	else
		g->files = f->next;
	
	if (f->next)
		f->next->prev = f->prev;
	
	MemChunkFree(f);
	
	return res;
}

Err FSFileClose(UInt16 fsLibRefNum, FileRef fileRef)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct FileState *f = fsPrvFindFile(g, fileRef);
	
	if (!f)
		return vfsErrFileBadRef;

	cachePrvClean(g, false);
	
	return mErrMap[fsPrvFileClose(g, f)];
}

Err FSFileRead(UInt16 fsLibRefNum, FileRef fileRef, UInt32 numBytes, void *bufBaseP, UInt32 offset, Boolean dataStoreBased, UInt32 *numBytesReadP)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct FileState *f = fsPrvFindFile(g, fileRef);
	UInt8 *dst = ((UInt8*)bufBaseP) + offset;
	UInt32 done, atOnce;
	FRESULT res;
	UINT now;
	
	if (!f)
		return vfsErrFileBadRef;
	
	if (f->isDir)
		return vfsErrIsADirectory;
	
	atOnce = dataStoreBased ? sizeof(g->dataWriteBuf) : OP_AT_ONCE;
	for (done = 0; done < numBytes; done += now) {
		
		now = numBytes - done;
		if (now > atOnce)
			now = atOnce;
		
		res = f_read(&f->f, dataStoreBased ? g->dataWriteBuf : dst, now, &now);
		if (res != FR_OK || !now)
			break;
		
		if (!dataStoreBased)
			dst += now;
		else {
		
			DmWrite(bufBaseP, offset, g->dataWriteBuf, now);
			offset += now;
		}
	}
	
	if (numBytesReadP)
		*numBytesReadP = done;

	if (res == FR_OK && done !=numBytes && f_eof(&f->f))	//only report EOF if we did not read the entire requested length.
		return vfsErrFileEOF;
	
	return mErrMap[res];
}

Err FSFileWrite(UInt16 fsLibRefNum, FileRef fileRef, UInt32 numBytes, const void *dataP, UInt32 *numBytesWrittenP)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct FileState *f = fsPrvFindFile(g, fileRef);
	const UInt8 *src = (const UInt8*)dataP;
	UInt32 done;
	FRESULT res;
	UINT now;
	
	if (!f)
		return vfsErrFileBadRef;
	
	if (f->isDir)
		return vfsErrIsADirectory;
	
	for (done = 0; done < numBytes; done += now, src += now) {
		
		now = numBytes - done;
		if (now > OP_AT_ONCE)
			now = OP_AT_ONCE;

		res = f_write(&f->f, src, now, &now);
		if (res != FR_OK || !now)
			break;
	}
	
	if (numBytesWrittenP)
		*numBytesWrittenP = done;
	
	//no cache clean here. we make no promises pon consistency if you do not close your file
	
	return mErrMap[res];
}

Err FSFileDelete(UInt16 fsLibRefNum, UInt16 volRefNum, const Char *pathNameP)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct VolumeState *v = fsPrvFindVolume(g, volRefNum);
	FRESULT res;
	
	if (!v)
		return vfsErrVolumeBadRef;
	
	res = f_unlink(&v->vol, pathNameP);

	cachePrvClean(g, false);
	
	return mErrMap[res];
}

Err FSFileRename(UInt16 fsLibRefNum, UInt16 volRefNum, const Char *pathNameP, const Char *newNameP)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct VolumeState *v = fsPrvFindVolume(g, volRefNum);
	Int32 ofst = 0, len, delimPos = -1;
	Char *newPath;
	FRESULT res;
	WChar chr;
	
	if (!v)
		return vfsErrVolumeBadRef;
	
	//we need a new complete path but PalmOS only gives name, so we need to MAKE the path
	//first find the last delimeter (ignoring if it is the last char, since FileZ
	// likes to pass a path ending in slash for dir renaming)
	len = StrLen(pathNameP);
	while (ofst < len) {
		
		UInt16 chrLen = TxtGetNextChar(pathNameP, ofst, &chr);
		if (chr == '/' && ofst + chrLen < len)
			delimPos = ofst;
		
		ofst += chrLen;
	}
	newPath = MemPtrNew(StrLen(newNameP) + delimPos + 2);	//big enough;
	if (!newPath)
		return memErrNotEnoughSpace;
	
	MemMove(newPath, pathNameP, delimPos + 1);
	StrCopy(newPath + delimPos + 1, newNameP);
	
	res = f_rename(&v->vol, pathNameP, newPath);
	MemPtrFree(newPath);

	cachePrvClean(g, false);
	
	return mErrMap[res];
}

Err FSFileSeek(UInt16 fsLibRefNum, FileRef fileRef, FileOrigin origin, Int32 offset)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct FileState *f = fsPrvFindFile(g, fileRef);
	Boolean pastOrAtEnd = false;
	FRESULT res;
	FSIZE_t to;
	
	if (!f)
		return vfsErrFileBadRef;
	
	if (f->isDir)
		return vfsErrIsADirectory;
	
	switch (origin) {
		case vfsOriginBeginning:
		default:
			if (offset < 0)
				return sysErrParamErr; 
			to = offset < 0 ? 0 : offset;
			break;
		
		case vfsOriginCurrent:
			if (offset < 0 && f_tell(&f->f) < (UInt32)-offset)
				to = 0;
			else if (offset > 0 && f_size(&f->f) - f_tell(&f->f) <= (UInt32)offset) {
				pastOrAtEnd = true;
				to = f_size(&f->f);
			}
			else
				to = f_tell(&f->f) + offset;
			break;
		
		case vfsOriginEnd:
			if (offset >= 0) {
				pastOrAtEnd = true;
				to = f_size(&f->f);
			}
			else
				to = f_size(&f->f) + offset;
			break;
	}
	
	res = f_lseek(&f->f, to);
	
	if (res == FR_OK && pastOrAtEnd)
		return vfsErrFileEOF;
		
	return mErrMap[res];
}

Err FSFileEOF(UInt16 fsLibRefNum, FileRef fileRef)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct FileState *f = fsPrvFindFile(g, fileRef);
	
	if (!f)
		return vfsErrFileBadRef;
	
	if (f->isDir)
		return vfsErrIsADirectory;
	
	return f_eof(&f->f) ? vfsErrFileEOF : errNone;
}

Err FSFileTell(UInt16 fsLibRefNum, FileRef fileRef, UInt32 *filePosP)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct FileState *f = fsPrvFindFile(g, fileRef);
	
	if (!f)
		return vfsErrFileBadRef;
	
	if (f->isDir)
		return vfsErrIsADirectory;
	
	*filePosP = f_tell(&f->f);
	
	return errNone;
}

Err FSFileResize(UInt16 fsLibRefNum, FileRef fileRef, UInt32 newSize)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct FileState *f = fsPrvFindFile(g, fileRef);
	FRESULT res, tr;
	FSIZE_t pos;
	
	if (!f)
		return vfsErrFileBadRef;
	
	if (f->isDir)
		return vfsErrIsADirectory;
	
	if (f_size(&f->f) == newSize)
		return errNone;
	
	pos = f_tell(&f->f);
	res = f_lseek(&f->f, newSize);
	if (res == FR_OK)
		res = f_truncate(&f->f);
	
	tr = f_lseek(&f->f, pos);
	if (res == FR_OK)
		res = tr;
	
	cachePrvClean(g, false);
	
	return mErrMap[res];
}

Err FSFileGetAttributes(UInt16 fsLibRefNum, FileRef fileRef, UInt32 *attributesP)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct FileState *f = fsPrvFindFile(g, fileRef);
	
	if (!f)
		return vfsErrFileBadRef;
	
	*attributesP = f->isDir ? f->d.obj.attr : f->f.obj.attr;
	
	return errNone;
}

Err FSFileSetAttributes(UInt16 fsLibRefNum, FileRef fileRef, UInt32 attributes)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct FileState *f = fsPrvFindFile(g, fileRef);
	FRESULT res;
	
	if (!f)
		return vfsErrFileBadRef;
	
	res = f_chmod_obj(f->isDir ? &f->d.obj : &f->f.obj, attributes, AM_RDO | AM_HID | AM_SYS | AM_ARC);
	
	cachePrvClean(g, false);
	
	return mErrMap[res];
}

static Err fsPrvDateOp(UInt16 fsLibRefNum, FileRef fileRef, UInt16 whichDate, UInt32 *dateP, Boolean isSet)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct FileState *f = fsPrvFindFile(g, fileRef);
	FRESULT res;
	UInt32 val;
	
	if (!f)
		return vfsErrFileBadRef;
	
	if (whichDate != vfsFileDateCreated && whichDate != vfsFileDateModified && whichDate != vfsFileDateAccessed)
		return expErrUnsupportedOperation;
	
	if (!dateP)
		return sysErrParamErr;
	
	val = fsPrvPalmTimeToFatTime(*dateP);
	res = f_time_obj(f->isDir ? &f->d.obj : &f->f.obj, isSet, (whichDate == vfsFileDateCreated) ? &val : NULL, (whichDate == vfsFileDateModified) ? &val : NULL, (whichDate == vfsFileDateAccessed) ? &val : NULL);
	*dateP = fsPrvFatTimeToPalmTime(val);
	
	return mErrMap[res];
}

Err FSFileGetDate(UInt16 fsLibRefNum, FileRef fileRef, UInt16 whichDate, UInt32 *dateP)
{
	return fsPrvDateOp(fsLibRefNum, fileRef, whichDate, dateP, false);
}

Err FSFileSetDate(UInt16 fsLibRefNum, FileRef fileRef, UInt16 whichDate, UInt32 date)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	Err e = fsPrvDateOp(fsLibRefNum, fileRef, whichDate, &date, true);
	
	cachePrvClean(g, false);
	
	return e;
}

Err FSFileSize(UInt16 fsLibRefNum, FileRef fileRef, UInt32 *fileSizeP)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct FileState *f = fsPrvFindFile(g, fileRef);
	
	if (!f)
		return vfsErrFileBadRef;
	
	if (f->isDir)
		return vfsErrIsADirectory;
	
	*fileSizeP = f_size(&f->f);
	
	return errNone;
}
 
Err FSDirCreate(UInt16 fsLibRefNum, UInt16 volRefNum, const Char *dirNameP)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct VolumeState *v = fsPrvFindVolume(g, volRefNum);
	FRESULT res;
	
	if (!v)
		return vfsErrVolumeBadRef;

	res = f_mkdir(&v->vol, dirNameP);

	cachePrvClean(g, false);
	
	return mErrMap[res];
}

Err FSDirEntryEnumerate(UInt16 fsLibRefNum, FileRef dirRef, UInt32 *dirEntryIteratorP, FileInfoType *infoP)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct FileState *f = fsPrvFindFile(g, dirRef);
	FRESULT res;
	
	if (!f)
		return vfsErrFileBadRef;
	
	if (!f->isDir)
		return vfsErrNotADirectory;

	if (*dirEntryIteratorP == vfsIteratorStart)
		f_rewinddir(&f->d);
	
	res = f_readdir(&f->d, &g->tmpFI);
	if (res != FR_OK)
		return mErrMap[res];
	
	if (!*g->tmpFI.fname) {
		
		*dirEntryIteratorP = vfsIteratorStop;
		return expErrEnumerationEmpty;
	}
	
	(*dirEntryIteratorP)++;
	
	if (infoP) {
		
		infoP->attributes = g->tmpFI.fattrib;
		if (infoP->nameP && infoP->nameBufLen) {
		
			StrNCopy(infoP->nameP, g->tmpFI.fname, infoP->nameBufLen - 1);
			infoP->nameP[infoP->nameBufLen - 1] = 0;
		}
	}
	
	return errNone;
}

Err FSVolumeFormat(UInt16 fsLibRefNum, VFSAnyMountParamPtr mountParamP)
{
	const VFSSlotMountParamType *mountParam = (VFSSlotMountParamType*)mountParamP;
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct VolumeState *v;
	MKFS_PARM mkfsParams;
	FRESULT res;
	
	MemSet(&mkfsParams, sizeof(MKFS_PARM), 0);
	mkfsParams.fmt = FM_FAT;
	mkfsParams.n_fat = 2;

	if (mountParamP->mountClass != vfsMountClass_SlotDriver)
		return expErrUnsupportedOperation;
	
	v = MemChunkNew(0, sizeof(struct VolumeState), memNewChunkFlagNonMovable);
	if (!v)
		return memErrNotEnoughSpace;
	MemSet(v, sizeof(struct VolumeState), 0);
	v->g = g;
	
	v->slotLibRef = mountParam->slotLibRefNum;
	v->slotRef = mountParam->slotRefNum;
	v->volRefNum = mountParam->vfsMountParam.volRefNum;
	
	//try fat12/16
	res = f_mkfs(v, &mkfsParams, g->dataWriteBuf, sizeof(g->dataWriteBuf));
	if (res == FR_MKFS_ABORTED) {	//try fat32
		mkfsParams.fmt |= FM_FAT32;
		res = f_mkfs(v, &mkfsParams, g->dataWriteBuf, sizeof(g->dataWriteBuf));
	}
	
	MemChunkFree(v);
	
	cachePrvClean(g, false);
	
	return mErrMap[res];
}

Err FSVolumeMount(UInt16 fsLibRefNum, VFSAnyMountParamPtr mountParamP)
{
	const VFSSlotMountParamType *mountParam = (VFSSlotMountParamType*)mountParamP;
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct VolumeState *v;
	FRESULT res;


	if (mountParamP->mountClass != vfsMountClass_SlotDriver)
		return expErrUnsupportedOperation;

	v = MemChunkNew(0, sizeof(struct VolumeState), memNewChunkFlagNonMovable);
	if (!v)
		return memErrNotEnoughSpace;
	MemSet(v, sizeof(struct VolumeState), 0);
	v->g = g;
	
	v->slotLibRef = mountParam->slotLibRefNum;
	v->slotRef = mountParam->slotRefNum;
	v->volRefNum = mountParam->vfsMountParam.volRefNum;
	
	res = f_mount(&v->vol, v);
	if (res != FR_OK)
		MemChunkFree(v);
	else {
		
		if (g->volumes)
			g->volumes->prev = v;
		v->next = g->volumes;
		v->prev = NULL;
		g->volumes = v;
	}
	
	return mErrMap[res];
}

Err FSVolumeUnmount(UInt16 fsLibRefNum, UInt16 volRefNum)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct VolumeState *v = fsPrvFindVolume(g, volRefNum);
	struct FileState *f = g->files;
	
	if (!v)
		return vfsErrVolumeBadRef;
	
	cachePrvClean(v->g, true);
	
	//unlink it
	
	if (v->prev)
		v->prev->next = v->next;
	else
		g->volumes = v->next;
	
	if (v->next)
		v->next->prev = v->prev;
	
	//close files
	while (f) {
		
		FATFS *filesVol = (f->isDir) ? f->f.obj.fs : f->f.obj.fs;
		struct FileState *nextFile = f->next;
		
		if (filesVol == &v->vol)
			(void)fsPrvFileClose(g, f);	//result does not matter as unmount cnanot report failure (breaks things)
		
		f = nextFile;
	}
	
	//done
	MemChunkFree(v);
	
	return errNone;
}

Err FSVolumeInfo(UInt16 fsLibRefNum, UInt16 volRefNum, VolumeInfoType *volInfoP)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct VolumeState *v = fsPrvFindVolume(g, volRefNum);
	ExpCardInfoType ci;
	Err e;
	
	if (!v)
		return vfsErrVolumeBadRef;

	volInfoP->fsType = vfsFilesystemType_VFAT;
	volInfoP->fsCreator = 'PpSF';
	volInfoP->mountClass = vfsMountClass_SlotDriver;
	volInfoP->slotLibRefNum = v->slotLibRef;
	volInfoP->slotRefNum = v->slotRef;
	volInfoP->reserved = 0;
	
	e = SlotCardMediaType(v->slotLibRef, v->slotRef, &volInfoP->mediaType);
	if (e != errNone)
		return e;
	
	e = SlotCardInfo(v->slotLibRef, v->slotRef, &ci);
	if (e != errNone)
		return e;
	
	volInfoP->attributes = vfsVolumeAttrSlotBased | 
		((ci.capabilityFlags & expCapabilityReadOnly) ? vfsVolumeAttrReadOnly : 0) |
		((ci.capabilityFlags & expCapabilityHidden) ? vfsVolumeAttrHidden : 0);
	
	return errNone;
}

Err FSVolumeGetLabel(UInt16 fsLibRefNum, UInt16 volRefNum, Char *labelP, UInt16 bufLen)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct VolumeState *v = fsPrvFindVolume(g, volRefNum);
	FRESULT res;
	
	if (!v)
		return vfsErrVolumeBadRef;
	
	res = f_getlabel(&v->vol, g->lblBuf);

	if (res != FR_OK)
		return mErrMap[res];
	
	if (bufLen) {
		StrNCopy(labelP, g->lblBuf, bufLen - 1);
		labelP[bufLen - 1] = 0;
	}
	
	return errNone;
}

Err FSVolumeSetLabel(UInt16 fsLibRefNum, UInt16 volRefNum, const Char *labelP)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct VolumeState *v = fsPrvFindVolume(g, volRefNum);
	FRESULT res;
	
	if (!v)
		return vfsErrVolumeBadRef;
	
	res = f_setlabel(&v->vol, labelP);
	
	cachePrvClean(g, false);
	
	return mErrMap[res];
}

static UInt32 fsPrvMulAndClip(UInt32 nClus, UInt32 clusSz)		//return MIN(MAX_UINT32, nClus * clusSz)
{
	uint64_t sz = (uint64_t)nClus * clusSz;
	
	if (sz >= 0x100000000ul)
		return 0xffffffff;
	else
		return sz;
}

Err FSVolumeSize(UInt16 fsLibRefNum, UInt16 volRefNum, UInt32 *volumeUsedP, UInt32 *volumeTotalP)
{
	struct Globals *g = fsPrvGetGlobals(fsLibRefNum);
	struct VolumeState *v = fsPrvFindVolume(g, volRefNum);
	UInt32 volSzInSecs, volFreeInSecs, volUsedInSecs, maxRepresentableSecs = 0x100000000ull / slotSectorSize - 1;
	DWORD freeClus;
	FRESULT res;
	
	if (!v)
		return vfsErrVolumeBadRef;
	
	res = f_getfree(&v->vol, &freeClus);
	if (res != FR_OK)
		return mErrMap[res];

//	clusterSize = (UInt32)slotSectorSize * v->vol.csize;
	
	//as we cnanot report over 4GB, err on the side of reporting accurate free space
	volSzInSecs = (v->vol.n_fatent - 2) * v->vol.csize;
	volFreeInSecs = freeClus * v->vol.csize;
	
	if (volSzInSecs > maxRepresentableSecs)
		volSzInSecs = maxRepresentableSecs;
	
	if (volFreeInSecs > maxRepresentableSecs)
		volFreeInSecs = maxRepresentableSecs;
	
	volUsedInSecs = volSzInSecs - volFreeInSecs;
	
	if (volumeUsedP)
		*volumeUsedP = volUsedInSecs * slotSectorSize;
	
	if (volumeTotalP)
		*volumeTotalP = volSzInSecs * slotSectorSize;
	
	return errNone;
}


#ifdef __GNUC__

	static void SlimFatDispatchTable(void)
	{
		asm volatile(
			"dc.w	194							\n\t"
			"dc.w	66							\n\t"
			"dc.w	70							\n\t"
			"dc.w	74							\n\t"
			"dc.w	78							\n\t"
			"dc.w	82							\n\t"
			"dc.w	86							\n\t"
			"dc.w	90							\n\t"
			"dc.w	94							\n\t"
			"dc.w	98							\n\t"
			"dc.w	102							\n\t"
			"dc.w	106							\n\t"
			"dc.w	110							\n\t"
			"dc.w	114							\n\t"
			"dc.w	118							\n\t"
			"dc.w	122							\n\t"
			"dc.w	126							\n\t"
			"dc.w	130							\n\t"
			"dc.w	134							\n\t"
			"dc.w	138							\n\t"
			"dc.w	142							\n\t"
			"dc.w	146							\n\t"
			"dc.w	150							\n\t"
			"dc.w	154							\n\t"
			"dc.w	158							\n\t"
			"dc.w	162							\n\t"
			"dc.w	166							\n\t"
			"dc.w	170							\n\t"
			"dc.w	174							\n\t"
			"dc.w	178							\n\t"
			"dc.w	182							\n\t"
			"dc.w	186							\n\t"
			"dc.w	190							\n\t"
			"jmp	FSLibOpen(%pc)				\n\t"
			"jmp	FSLibClose(%pc)				\n\t"
			"jmp	FSLibSleep(%pc)				\n\t"
			"jmp	FSLibWake(%pc)				\n\t"
			"jmp	FSLibAPIVersion(%pc)		\n\t"
			"jmp	FSCustomControl(%pc)		\n\t"
			"jmp	FSFilesystemType(%pc)		\n\t"
			"jmp	FSFileCreate(%pc)			\n\t"
			"jmp	FSFileOpen(%pc)				\n\t"
			"jmp	FSFileClose(%pc)			\n\t"
			"jmp	FSFileRead(%pc)				\n\t"
			"jmp	FSFileWrite(%pc)			\n\t"
			"jmp	FSFileDelete(%pc)			\n\t"
			"jmp	FSFileRename(%pc)			\n\t"
			"jmp	FSFileSeek(%pc)				\n\t"
			"jmp	FSFileEOF(%pc)				\n\t"
			"jmp	FSFileTell(%pc)				\n\t"
			"jmp	FSFileResize(%pc)			\n\t"
			"jmp	FSFileGetAttributes(%pc)	\n\t"
			"jmp	FSFileSetAttributes(%pc)	\n\t"
			"jmp	FSFileGetDate(%pc)			\n\t"
			"jmp	FSFileSetDate(%pc)			\n\t"
			"jmp	FSFileSize(%pc)				\n\t"
			"jmp	FSDirCreate(%pc)			\n\t"
			"jmp	FSDirEntryEnumerate(%pc)	\n\t"
			"jmp	FSVolumeFormat(%pc)			\n\t"
			"jmp	FSVolumeMount(%pc)			\n\t"
			"jmp	FSVolumeUnmount(%pc)		\n\t"
			"jmp	FSVolumeInfo(%pc)			\n\t"
			"jmp	FSVolumeGetLabel(%pc)		\n\t"
			"jmp	FSVolumeSetLabel(%pc)		\n\t"
			"jmp	FSVolumeSize(%pc)			\n\t"
			".ascii   \"SlimFat\\0\"			\n\t"
		);
	}

#else
	
	#define numberOfAPIs        		32
	#define libDispatchEntry(index)		(2 * (1 + numberOfAPIs) + index * 4)

	static void asm SlimFatDispatchTable(void)
	{
		/* Offset to library name */
		DC.W		@Name
		
	    /* entry point offsets */
		DC.W		libDispatchEntry(0)
		DC.W		libDispatchEntry(1)
		DC.W		libDispatchEntry(2)
		DC.W		libDispatchEntry(3)
		DC.W		libDispatchEntry(4)
		DC.W		libDispatchEntry(5)
		DC.W		libDispatchEntry(6)
		DC.W		libDispatchEntry(7)
		DC.W		libDispatchEntry(8)
		DC.W		libDispatchEntry(9)
		DC.W		libDispatchEntry(10)
		DC.W		libDispatchEntry(11)
		DC.W		libDispatchEntry(12)
		DC.W		libDispatchEntry(13)
		DC.W		libDispatchEntry(14)
		DC.W		libDispatchEntry(15)
		DC.W		libDispatchEntry(16)
		DC.W		libDispatchEntry(17)
		DC.W		libDispatchEntry(18)
		DC.W		libDispatchEntry(19)
		DC.W		libDispatchEntry(20)
		DC.W		libDispatchEntry(21)
		DC.W		libDispatchEntry(22)
		DC.W		libDispatchEntry(23)
		DC.W		libDispatchEntry(24)
		DC.W		libDispatchEntry(25)
		DC.W		libDispatchEntry(26)
		DC.W		libDispatchEntry(27)
		DC.W		libDispatchEntry(28)
		DC.W		libDispatchEntry(29)
		DC.W		libDispatchEntry(30)
		DC.W		libDispatchEntry(31)
	
	
	    /* entry jumps */
        JMP     FSLibOpen
        JMP     FSLibClose
        JMP     FSLibSleep
        JMP     FSLibWake
        JMP     FSLibAPIVersion
        JMP     FSCustomControl
        JMP     FSFilesystemType
        JMP     FSFileCreate
        JMP     FSFileOpen
        JMP     FSFileClose
        JMP     FSFileRead
        JMP     FSFileWrite
        JMP     FSFileDelete
        JMP     FSFileRename
        JMP     FSFileSeek
        JMP     FSFileEOF
        JMP     FSFileTell
        JMP     FSFileResize
        JMP     FSFileGetAttributes
        JMP     FSFileSetAttributes
        JMP     FSFileGetDate
        JMP     FSFileSetDate
        JMP     FSFileSize
        JMP     FSDirCreate
        JMP     FSDirEntryEnumerate
        JMP     FSVolumeFormat
        JMP     FSVolumeMount
        JMP     FSVolumeUnmount
        JMP     FSVolumeInfo
        JMP     FSVolumeGetLabel
        JMP     FSVolumeSetLabel
        JMP     FSVolumeSize
	@Name:
		DC.B	"SlimFat"
	}
#endif


Err __Startup__(UInt16 refNum, SysLibTblEntryType *entryP)
{
	entryP->dispatchTblP = (void**)&SlimFatDispatchTable;
	entryP->globalsP = NULL;
	return 0;
}


//text

static CharEncodingType fsPrvGetDeviceCharEncoding(void)
{
	UInt32 val;
	
	if (errNone == FtrGet(sysFtrCreator, sysFtrNumEncoding, &val))
		return (CharEncodingType)val;
	
	return charEncodingPalmLatin;
}

DWORD tchar2uni(const Char **str)
{
	UInt16 len;
	WChar chr;
	
	len = TxtGetNextChar(*str, 0, &chr);
	(*str) += len;
	
	return chr;
}

UINT put_utf(DWORD chr,	TCHAR *buf,	UINT szb)
{
	if (TxtCharSize(chr) > szb)
		return 0;
	
	return TxtSetNextChar(buf, 0, chr);
}


WCHAR ff_oem2uni(WCHAR oem)		//wc is a byte, or (first << 8 + second) 
{
	Char bytes[2];
	UInt16 ret[3] /* because bug */;
	UInt16 srcLen, dstLen = sizeof(ret);
	
	if (oem >= 0x100) {
		
		bytes[0] = oem >> 8;
		bytes[1] = oem;
		srcLen = 2;
	}
	else {
		
		bytes[0] = oem;
		srcLen = 1;
	}
	
	return (errNone == TxtConvertEncoding(true, NULL, bytes, &srcLen, fsPrvGetDeviceCharEncoding(), (Char*)ret, &dstLen, charEncodingUCS2, NULL, 0) && dstLen == sizeof(UInt16)) ? ret[0] : 0;
}

WCHAR ff_uni2oem(DWORD uni)	//return is a byte, or (first << 8 + second) 
{
	Char ret[3];
	WChar wch = uni;
	UInt16 dstLen = sizeof(ret), srcLen = sizeof(wch);
	
	if (errNone != TxtConvertEncoding(true, NULL, (Char*)&wch, &srcLen, charEncodingUCS2, ret, &dstLen, fsPrvGetDeviceCharEncoding(), NULL, 0))
		return 0;
	
	if (dstLen == 1)
		return (UInt8)ret[0];
	
	if (dstLen != 2)
		return 0;
	
	return (((UInt16)(UInt8)ret[0]) << 8) + (UInt8)ret[1];
}



//time

DWORD get_fattime(void)
{
	return fsPrvPalmTimeToFatTime(TimGetSeconds());
}

//disk


static Err fsPrvGetNumSec(UInt16 slotLibRefNum, UInt16 slotRefNum, LBA_t *nSecP)
{
	union {
		struct CardMetricsTag v2;
		struct CardMetricsTypeV1 v1;
	} nfo;
	Err e;
	
	e = SlotCardMetrics(slotLibRefNum, slotRefNum, &nfo.v2);
	if (e != errNone)
		return e;
	
	if (isSonyEarlyVFS()) {
		
		if (nfo.v1.bytesPerSector != slotSectorSize)
			return sysErrParamErr;
		
		*nSecP = nfo.v1.totalSectors;
		MemPtrFree(nfo.v1.pl);
	}
	else {
		
		if (nfo.v2.bytesPerSector != slotSectorSize)
			return sysErrParamErr;
		
		*nSecP = nfo.v2.totalSectors;
	}
	return errNone;
}

DSTATUS disk_initialize(void *pdrvp)
{
	return disk_status(pdrvp);
}

DSTATUS disk_status(void *pdrvp)
{
	struct VolumeState *v = (struct VolumeState*)pdrvp;
	Err e;
	
	e = ExpCardInfo(v->slotRef, &v->g->ci);
	if (e != errNone)
		return STA_NOINIT;
	
	if (!(v->g->ci.capabilityFlags & (expCapabilityHasStorage)))
		return STA_NODISK;
	
	return (v->g->ci.capabilityFlags & expCapabilityReadOnly) ? STA_PROTECT : 0;
}


#if CACHE_NUM_SECTORS
	
	static const char mCacheDbName[] = "SlimFatCache";
	
	static void cachePrvInit(struct Globals *g)
	{
		MemHandle mh;
		DmOpenRef db;
		LocalID LID;
		UInt16 i;
		Err e;

		g->numCacheEntries = 0;
		
		LID = DmFindDatabase(0, mCacheDbName);
		if (LID) {
		
			e = DmDeleteDatabase(0, LID);
			if (e != errNone)
				return;
		}
	
		e = DmCreateDatabase(0, mCacheDbName, MY_CRID, 'vfs$', true);
		if (e != errNone)
			return;
		
		LID = DmFindDatabase(0, mCacheDbName);
		if (!LID)
			return;
		
		db = DmOpenDatabase(0, LID, dmModeReadWrite);
		if (e != errNone)
			return;
		
		for (i = 0; i < CACHE_NUM_SECTORS; i++) {
			
			mh = DmNewResource(db, 'data', i, slotSectorSize);
			if (!mh)
				break;
			
			g->cacheEntries[i].dataH = mh;
			g->cacheEntries[i].valid = false;
		}
		g->numCacheEntries = i;
		
		//now link them into a linked list
		if (g->numCacheEntries) {
		
			for (i = 0; i < g->numCacheEntries; i++) {
				
				g->cacheEntries[i].prev = &g->cacheEntries[i - 1];
				g->cacheEntries[i].next = &g->cacheEntries[i + 1];
			}
			g->cacheEntries[0].prev = NULL;
			g->cacheEntries[g->numCacheEntries - 1].next = NULL;
			
			g->head = &g->cacheEntries[0];
			g->tail = &g->cacheEntries[g->numCacheEntries - 1];
		}
	}
	
	static void cachePrvEnqueueEntry(struct VolumeState *v, struct CacheEntry *ce)
	{
		struct Globals *g = v->g;
		
		//set volume info
		ce->slotLibRef = v->slotLibRef;
		ce->slotRef = v->slotRef;
		
		//skip if possible
		if (ce == g->head)
			return;
		
		//remove (rememmber we're not first)
		ce->prev->next = ce->next;
		if (ce->next)
			ce->next->prev = ce->prev;
		else
			g->tail = ce->prev;
		
		//insert (we are guaranteed to become first)
		ce->prev = NULL;
		ce->next = g->head;	
		if (ce->next)
			ce->next->prev = ce;
		else
			g->tail = ce;
		g->head = ce;
	}
	
	static void cachePrvCleanEntry(struct CacheEntry *ce)
	{
		UInt32 nSec = 1;
		MemHandle mh;
		
		if (ce->valid && ce->dirty) {
		
			(void)SlotCardSectorWrite(ce->slotLibRef, ce->slotRef, ce->sector, MemHandleLock(mh = ce->dataH), &nSec);
			MemHandleUnlock(mh);
			ce->dirty = false;
		}
	}
	
	static struct CacheEntry* cachePrvGetEntry(struct VolumeState *v)
	{
		struct CacheEntry *ce = v->g->tail;
		
		if (ce)
			cachePrvCleanEntry(ce);
		
		return ce;
	}
	
	static void cachePrvClean(struct Globals *g, Boolean andInvalidate)
	{
		struct CacheEntry *ce;
		
		for (ce = g->head; ce; ce = ce->next) {
			
			cachePrvCleanEntry(ce);
			if (andInvalidate)
				ce->valid = false;
		}
	}
	
	static struct CacheEntry* cachePrvFind(struct VolumeState *v, LBA_t sector)
	{
		struct CacheEntry *ce;
		
		for (ce = v->g->head; ce; ce = ce->next) {
			
			if (ce->valid && ce->sector == sector && ce->slotLibRef == v->slotLibRef && ce->slotRef == v->slotRef) 	//match? read from it and move to head
				return ce;
		}
		
		return NULL;
	}

	//we read-allocate
	DRESULT disk_read(void *pdrvp, BYTE *buff, LBA_t sector, UINT count)
	{
		struct VolumeState *v = (struct VolumeState*)pdrvp;
		struct CacheEntry *ce = v->g->cacheEntries;
		UInt32 nSec = 1;
		MemHandle mh;
		UINT secIdx;
		Err e;
		
		for (secIdx = 0; secIdx < count; secIdx++, sector++, buff += slotSectorSize) {
			
			ce = cachePrvFind(v, sector);
			if (ce) {						//found
			
				MemMove(buff, MemHandleLock(mh = ce->dataH), slotSectorSize);
				MemHandleUnlock(mh);
				cachePrvEnqueueEntry(v, ce);
			}
			else {							//not found in cache - fill it (we read-allocate)
				
				ce = cachePrvGetEntry(v);
				e = SlotCardSectorRead(v->slotLibRef, v->slotRef, sector, buff, &nSec);
				if (e != errNone || nSec != 1)
					return RES_ERROR;
				
				if (ce) {
					DmWrite(MemHandleLock(mh = ce->dataH), 0, buff, slotSectorSize);
					ce->sector = sector;
					ce->valid = true;
					MemHandleUnlock(mh);
					cachePrvEnqueueEntry(v, ce);
				}
			}
		}
		
		return RES_OK;
	}
	
	//we do not write-allocate
	DRESULT disk_write(void *pdrvp, const BYTE* buff, LBA_t sector, UINT count)
	{
		struct VolumeState *v = (struct VolumeState*)pdrvp;
		struct CacheEntry *ce = v->g->cacheEntries;
		UInt32 nSec = 1;
		MemHandle mh;
		UINT secIdx;
		Err e;
		
		for (secIdx = 0; secIdx < count; secIdx++, sector++, buff += slotSectorSize) {
		
			ce = cachePrvFind(v, sector);
			
			if (ce) {						//found
				
				DmWrite(MemHandleLock(mh = ce->dataH), 0, buff, slotSectorSize);
				ce->dirty = true;
				MemHandleUnlock(mh);
				cachePrvEnqueueEntry(v, ce);
			}
			else {							//not found in cache - just write
				
				e = SlotCardSectorWrite(v->slotLibRef, v->slotRef, sector, (UInt8*)buff, &nSec);
				if (e != errNone || nSec != 1)
					return RES_ERROR;
			}
		}
		
		return RES_OK;
	}

#else

	static void cachePrvInit(struct Globals *g)
	{
		//nothing
	}
	
	static void cachePrvClean(struct Globals *g, Boolean andInvalidate)
	{
		//nothing
	}

	DRESULT disk_read(void *pdrvp, BYTE* buff, LBA_t sector, UINT count)
	{
		struct VolumeState *v = (struct VolumeState*)pdrvp;
		UInt32 nSec = count;
		Err e;
		
		e = SlotCardSectorRead(v->slotLibRef, v->slotRef, sector, (UInt8*)buff, &nSec);
		
		return (e == errNone && nSec == count) ? RES_OK : RES_ERROR;
	}

	DRESULT disk_write(void *pdrvp, const BYTE* buff, LBA_t sector, UINT count)
	{
		struct VolumeState *v = (struct VolumeState*)pdrvp;
		UInt32 nSec = count;
		Err e;
		
		e = SlotCardSectorWrite(v->slotLibRef, v->slotRef, sector, (UInt8*)buff, &nSec);
		if (e == expErrCardReadOnly)
			return RES_WRPRT;
		
		return (e == errNone && nSec == count) ? RES_OK : RES_ERROR;
	}

#endif


DRESULT disk_ioctl(void *pdrvp, BYTE cmd, void* buff)
{
	struct VolumeState *v = (struct VolumeState*)pdrvp;
	 
	switch (cmd) {
		case CTRL_SYNC:
			cachePrvClean(v->g, false);
			break;

		case GET_SECTOR_COUNT:
			if (errNone == fsPrvGetNumSec(v->slotLibRef, v->slotRef, (LBA_t*)buff))
				break;
			//fallthrough
		
		default:
			return RES_ERROR;
	}
	
	return RES_OK;
}







