#include "printf.h"
#include "prc.h"

//assumes LE host

#define MAX_RESOURCES						32

#define PALM_DB_ATTR_RES_DB					0x0001	// is PRC? (else PDB)
#define PALM_DB_ATTR_INVALID_FLAGS			0x7000


struct PrcRes {		//these are stored in order, so size can be inferred from offset of next res
	uint32_t type;
	uint16_t id;
	uint32_t ofst;	//from file start
} __attribute__((packed));

static struct PrcRes mRess[MAX_RESOURCES];
static uint32_t mFileSize;
static uint16_t mCurResPos;
static int16_t mNumRes, mCurRes;
static PrcFileReadF mReadF;
static void *mReadD;


bool prcOpen(PrcFileReadF readF, void *readD, uint32_t size, char *name, uint16_t *attrP, uint16_t *verP, uint32_t *typeP, uint32_t *cridP, uint16_t *numResP)
{
	uint32_t v32, curPos;
	uint_fast16_t i;
	uint16_t v16;
	
	if (mFileSize)		//only if we closed the last one ...
		return false;
	
	if (size < 0x4e) {	//min size
		printf("file too small at %u\n", size);
		return false;
	}
	
	if (PRC_NAME_BUF_SZ != readF(readD, name, PRC_NAME_BUF_SZ)) {
		printf("failed to read name\n");
		return false;
	}
	
	//check name for validity
	for (i = 0; i < PRC_NAME_BUF_SZ && name[i]; i++);
	if (i == PRC_NAME_BUF_SZ) {
		printf("name is invalid\n");
		return false;
	}
	
	//get and check attrs
	if (sizeof(v16) != readF(readD, &v16, sizeof(v16))) {
		printf("failed to read attrs\n");
		return false;
	}
	*attrP = v16 = __builtin_bswap16(v16);
	if (!(v16 & PALM_DB_ATTR_RES_DB)) {
		printf("! res db\n");
		return false;
	}
	if (v16 & PALM_DB_ATTR_INVALID_FLAGS) {
		printf("invalid flags 0x%04\n", v16);
		return false;
	}
	
	//get version
	if (sizeof(v16) != readF(readD, &v16, sizeof(v16))) {
		printf("failed to read version\n");
		return false;
	}
	*verP = __builtin_bswap16(v16);
	
	//skip things
	for (i = 0x24; i < 0x34; i += 2) {
		if (sizeof(v16) != readF(readD, &v16, sizeof(v16))) {
			printf("failed to skip1\n");
			return false;
		}
	}
	
	//app info & sort info ptrs
	for (i = 0; i < 2; i++) {
		if (sizeof(v32) != readF(readD, &v32, sizeof(v32))) {
			printf("failed to app/sort info\n");
			return false;
		}
		if (v32) {
			printf("app/sort info exists\n");
			return false;
		}
	}
	
	//read type and creator
	for (i = 0; i < 2; i++, typeP = cridP) {
		if (sizeof(v32) != readF(readD, &v32, sizeof(v32))) {
			printf("failed to type/crid\n");
			return false;
		}
		if (!v32) {
			printf("type/crid zero\n");
			return false;
		}
		*typeP = __builtin_bswap32(v32);
	}
	
	//skip seed, read next record list ptr
	for (i = 0; i < 2; i++) {
		if (sizeof(v32) != readF(readD, &v32, sizeof(v32))) {
			printf("failed to skip2\n");
			return false;
		}
	}
	if (v32) {	// no segmented record lists plz
		printf("segmented record lists not supported\n");
		return false;
	}
	
	//read number of resources & check it
	if (sizeof(v16) != readF(readD, &v16, sizeof(v16))) {
		printf("failed to read num res\n");
		return false;
	}
	*numResP = mNumRes = __builtin_bswap16(v16);
	if (mNumRes > MAX_RESOURCES) {
		printf("too many res at %u\n", mNumRes);
		return false;
	}
	
	for (i = 0; i < mNumRes; i++) {
		
		if (sizeof(struct PrcRes) != readF(readD, mRess + i, sizeof(struct PrcRes))) {
			printf("failed to read resInfo[%u]\n", i);
			return false;
		}
		
		mRess[i].type = __builtin_bswap32(mRess[i].type);
		mRess[i].id = __builtin_bswap16(mRess[i].id);
		mRess[i].ofst = __builtin_bswap32(mRess[i].ofst);
		
		if (mRess[i].ofst > size){
			printf("resInfo[%u].ofst == 0x%04x > size == 0x%04x\n", i, mRess[i].ofst, size);
			return false;
		}
		
		if (i && mRess[i - 1].ofst >= mRess[i].ofst) {
			printf("resInfo[%u].ofst == 0x%04x > resInfo[%u - 1].ofst == 0x%04x\n", i, mRess[i].ofst, i, mRess[i - 1].ofst);
			return false;
		}
	}
	curPos = 0x4e + mNumRes * sizeof(struct PrcRes);
	if (mRess[0].ofst < curPos) {
		printf("trying to rewind from %u to %u\n", curPos, mRess[0].ofst);
		return false;
	}
	
	if (mNumRes) {
	
		//fast forward to data
		while (curPos < mRess[0].ofst) {
			
			if (1 != readF(readD, &v16, 1)) {
				printf("failed to skip 3");
				return false;
			}
			curPos++;
		}
	}
	
	//we've opened it!
	mReadF = readF;
	mReadD = readD;
	mFileSize = size;
	mCurRes = -1;
	mCurResPos = 0;
	
	return true;
}

static uint32_t prcPrvGetResSz(int_fast16_t idx)
{
	if (idx == mNumRes - 1)
		return mFileSize - mRess[idx].ofst;
	else
		return mRess[idx + 1].ofst - mRess[idx].ofst;
}

bool prcBeginResRead(uint32_t *resTypeP, uint16_t *resIDP, uint32_t *resSizeP)
{
	uint32_t resSz;
	
	//make sure we are done with the last one, if any
	if (mCurRes >= 0 && mCurResPos < prcPrvGetResSz(mCurRes))
		return false;
	
	//make sure there IS a next one
	if (mCurRes == mNumRes - 1)
		return false;
	
	mCurRes++;
	mCurResPos = 0;
	*resTypeP = mRess[mCurRes].type;
	*resIDP = mRess[mCurRes].id;
	*resSizeP = resSz = prcPrvGetResSz(mCurRes);
	
	if (resSz >> 16)	//not supported
		return false;
	
	return true;
}

uint32_t prcContinueResRead(void *bufP, uint32_t space)
{
	uint8_t *buf = (uint8_t*)bufP;
	uint32_t avail, done, now;
	
	//make sure we have a current one
	if (mCurRes < 0)
		return 0;
	
	//make sure we have bytes to give
	avail = prcPrvGetResSz(mCurRes) - mCurResPos;
	if (avail > space)
		avail = space;
	
	//do
	for (done = 0; done < avail; done += now, mCurResPos += now, buf += now) {
		
		now = mReadF(mReadD, buf, avail - done);
		if (!now)
			break;
		
		printf("read resource at offset %u, %u left, now = %u\n", mCurResPos, avail - done, now);
		
		{
			uint_fast16_t i;
			
			buf = (uint8_t*)bufP;
			for (i = 0; i < done; i++) {
				if (!(i & 15))
					printf("\n%03x ", i);
				printf(" %02x", buf[i]);
			}
			printf("\n");
		}
	}
	
	return done;
}

void prcClose(void)
{
	mFileSize = 0;
}
