#include "timebase.h"
#include "pinmux.h"
#include "lpc802.h"
#include "uart.h"

#define UART_MAX_FRAME_ERRORS	4

static volatile uint8_t mUartRxBuf[64];
static uint16_t mUartRxBufW, mUartRxBufR;		//equality = empty
static uint8_t mNumFrameErrors;

void __attribute__((used)) UART0_IRQHandler(void)
{
	uint32_t val = USART0->RXDATSTAT;
	uint_fast16_t nextW = mUartRxBufW + 1;
	
	if (val & (3 << 13)) {	//skip shit, but count it
		
		if (!++mNumFrameErrors)	//cap it at max
			mNumFrameErrors--;
		
		USART0->STAT = 3 << 13;		
		return;
	}
	
	if (nextW == sizeof(mUartRxBuf))
    	nextW = 0;
	
    if (nextW != mUartRxBufR) {	//drop new data
    	
    	mUartRxBuf[mUartRxBufW] = val;
    	mUartRxBufW = nextW;
    }
}

void uartSetBaudrate(uint32_t baudrate)
{
	uint32_t divInt, divFrac;
	
	baudrate *= 16;	//because uart...
	
	//we have a fractional divider and an integer divider. integer divider can produce division of 1...65536
	//fractional divider can produce divisions of 1...2 and is optional
	
	divInt = CLOCKRATE / baudrate;
	divFrac = (CLOCKRATE / divInt) * 256 / baudrate - 256;
	
	SYSCON->FRG[0].FRGDIV = 0xff;
	SYSCON->FRG[0].FRGMULT = divFrac;
	USART0->BRG = divInt - 1;
}

void uartInit(void)
{
	SYSCON->UART0CLKSEL = 2;	//FRG drives the UART0 clock
	IOCON->PIO[mIoconRegOrder[PIN_NO_UART_TX]] = 0x20;
	IOCON->PIO[mIoconRegOrder[PIN_NO_UART_RX]] = 0x20;
	IOCON->PIO[mIoconRegOrder[PIN_NO_UART_RTS]] = 0x20;
	
	GPIO->DIRCLR[0] = (1 << PIN_NO_UART_RX) | (1 << PIN_NO_UART_RTS);
	GPIO->DIRSET[0] = (1 << PIN_NO_UART_TX);
	
	SWM0->PINASSIGN.PINASSIGN0 = (SWM0->PINASSIGN.PINASSIGN0 &~ 0x0000ffff) | (PIN_NO_UART_TX) | ((PIN_NO_UART_RX) << 8);
	USART0->CFG = 0;		//off
	uartSetBaudrate(9600);
	USART0->CTL = 0x0000;
	USART0->CFG = 0x00c00005;	//8n1, on, rx and tx inverted
	USART0->INTENSET = (1 << 0) | (1 << 13);
	timebaseDelay(CLOCKRATE / 1000);
	NVIC_SetPriority(USART0_IRQn, 1);
	NVIC_EnableIRQ(USART0_IRQn);
}

uint32_t uartRx(void *dstP, uint32_t nBytes, uint64_t endTime)
{
	uint8_t *dst = (uint8_t*)dstP;
	while (nBytes) {
		
		uint_fast16_t nextR = mUartRxBufR + 1;
		uint8_t val;
		
		if (nextR == sizeof(mUartRxBuf))
			nextR = 0;
		
		//this construct is thus for atomic test of buffer emptiness (and sleep if so) without missing the condition
		while(1) {
		
			if (uartSeenFrameErrs())
				return 0;
		
			asm volatile("":::"memory");
			if (mUartRxBufW != mUartRxBufR)
				break;
			
			asm volatile("wfi":::"memory");
			
			if (timebaseGet() > endTime)
				goto out;
		}
		
		val = mUartRxBuf[mUartRxBufR];
		mUartRxBufR = nextR;
		
		*dst++ = val;
		nBytes--;
	}
	
out:

	return dst - (uint8_t*)dstP;
}

void uartTxWait(void)
{
	while (!(USART0->STAT & 8));
}

void uartTx(const void *srcP, uint32_t len)
{
	const uint8_t *src = (const uint8_t*)srcP;
	while (len--) {
		
		//to make RTS work perfectly we need to wait for IDLE here (bit mask 8) but this slows down our normal xfers
		//instead we can allow up to one extra char to go out while CTS is indicating "stop". this is faster in
		// general and seems to work. This uses a test of "rx avail" mask 0x04
		
		while (!(USART0->STAT & 8));			//wait for idle (needed for CTS to sto soon enough)
		while (!GPIO->B[0][PIN_NO_UART_RTS]);	//wait for RTS
		USART0->TXDAT = *src++;
	}
}

bool uartSeenFrameErrs(void)
{
	return mNumFrameErrors > UART_MAX_FRAME_ERRORS;
}

void uartResetFrameErrors(void)
{
	mNumFrameErrors = 0;
}